#!/usr/bin/perl
###############################################################################
#    The Web Bin Packer
#    Copyright (c) 2004 Arne Bochem <binpacker@arneb.de>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
###############################################################################

use strict;
use CGI qw(:param);
use Fcntl qw(:flock);

###############################################################################
# Start of configuration. #####################################################

# Save data that people input.
my $be_nosy = 1;

# Name of nosy log file.
my $data_file = 'snoopy.log';

# End of configuration. #######################################################
###############################################################################

# This source file is inteded to be edited in 80x25 with a tab width of two
# spaces. Or something like that...

my $content_type = 'text/html';
($ENV{'HTTP_ACCEPT'} =~ m#application/xhtml\+xml#)
	&& ($content_type = 'application/xhtml+xml');

print "Content-Type: $content_type; charset=iso-8859-1\n\n";

my ($result, $value_pos, $container_size, $input, @whitespace)
	= ('', '', '', '', 'checked="checked" ', '', '');
(param('calc') eq 'nifty')
	&& (($result, $value_pos, $container_size, $input, @whitespace)
		= calculateResults($be_nosy, $data_file));

print <<EOF;
<?xml version="1.0" encoding="iso-8859-1"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">

<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
	<head>
		<meta name="robots" content="index, follow" />
		<title>The Web Bin Packer</title>
	</head>
	<body>
		<h1>The Web Bin Packer</h1>
		<h2>About</h2>
		<h3>What is this?</h3>
		<p>
			This is a bin packer. It uses <em>sorted packing</em> and the <em>first fit</em> algorithm. It can be used to find a method of distributing chunks of varying size over a number of containers of a specific size. The worst case result is never suboptimal by more than <em>22%</em>, which is probably the best you can hope for without using brute force.
		</p>
		<h3>What about the data I enter?</h3>
		<p>
			Do not trust <em>The Web Bin Packer</em> not to save your data for manual analysis or out of curiosity. It probably will save it. If you don't like this, you can download the source code (<em>The Web Bin Packer</em> is <em><a href="http://www.gnu.org">Free Software</a></em>) from <a href="http://www.arneb.de/software">the software section of <em>arneb.de</em></a> and run your own. You'll need a webserver (such as <a href="http://httpd.apache.org/"><em>The Apache HTTP Server</em></a>) with <a href="http://www.perl.org/"><em>Perl</em></a> support, though.
		</p>
		<h3>Give me an example, how to use this thing.</h3>
		<p>
			Ok. Example:
		</p>
		<ul>
			<li>Set the "Field separation" "Type" to "Use single whitespaces as field separators".</li>
			<li>Set "Field position" to 1 (2 - 1, because the first index is 0).</li>
			<li>Set the "Size of containers" to 10.</li>
			<li>
				Use the following as the input for "Chunk list":
				<p>
					Size: 16 Description: This thing is too large.<br />
					Size: 2 Description: This thing is quite small.<br />
					Size: -5 Description: This thing would like to be a hole, but it counts more like a zero.<br />
					Size: 8 Desc: Let's break the Convention.<br />
					This one won't work; actually it will, but it won't take up space.<br />
					Foo: 4 More on the small side.<br />
					And 7 knights take up space too.<br />
					But 4 more take up less space than 7.<br />
					Now 3 others join.
				</p>
			</li>
			<li>Try it out. You will see what happens.</li>
		</ul>
		<h2>Input</h2>
		<form action="bpacker.pl" method="post">
			<div>
				<input type="hidden" name="calc" value="nifty" />
				<h3>Field separation</h3>
				<h4>Type</h4>
				<ul>
					<li><label><input type="radio" name="whitespace" value="0" $whitespace[0]/> Use tabs as field separators.</label></li>
					<li><label><input type="radio" name="whitespace" value="1" $whitespace[1]/> Use multiple whitespaces as field separators.</label></li>
					<li><label><input type="radio" name="whitespace" value="2" $whitespace[2]/> Use single whitespaces as field separators.</label></li>
				</ul>
				<h4>Field position</h4>
				<input type="text" name="value_pos" $value_pos/> Position (starting with 0) of chunk size.
				<h3>Containers</h3>
				<input type="text" name="container_size" $container_size/> Size of containers.
				<h3>Chunk list</h3>
				<textarea rows="25" cols="80" name="input">$input</textarea>
				<h3>Form submission</h3>
				<input type="submit" value="Calculate chunk distribution." />
			</div>
		</form>$result
	</body>
</html>
EOF

sub makeNumber ($)
{
	my $number = shift;
	
	($number eq abs(int($number))) || return();
	return('value="'.int($number).'" ');
}

sub quoteCode ($)
{
	my $input = shift;
	$input =~ s/\&/&amp;/g;
	$input =~ s/</&lt;/g;
	$input =~ s/>/&gt;/g;
	return($input);
}

sub makeArray ($)
{
	my $whitespace = int(shift);
	my @whitespaces = ('', '', '');
	
	for (my $i = 0; $i < 3; $i++)
	{
		($i == $whitespace) && ($whitespaces[$i] = 'checked="checked" ');
	}
	return(@whitespaces);
}

sub makeItem ($$$)
{
	my ($line, $regexp, $value_pos) = @_;
	my @items = split(/$regexp/, $line);
	return([int($items[$value_pos]), $line]);
}

sub getContainerValue ($)
{
	my $container = shift;
	my $value;
	
	foreach my $item (@{$container})
	{
		$value += $item->[0];
	}

	return($value);
}

sub calculateResults
{
	my ($be_nosy, $data_file) = @_;
	my ($result, $value_pos, $container_size, $input, @whitespace)
		= ('', abs(int(param('value_pos'))), abs(int(param('container_size'))),
			param('input'), makeArray(param('whitespace')));

	my @regexps = ('\t', '\s+', '\s');
	my $ws = int(param('whitespace'));
	my $regexp = $regexps[($ws > -1 && $ws < 3) ? $ws : 0];

	# Get rid of problems.
	$input =~ s/\r\n/\n/g;
	$input =~ s/\r/\n/g;
	$input =~ s/\x00//g;
	$input =~ s/\x01//g;

	if ($be_nosy)
	{
		open(NOSY, '>>'.$data_file);
		flock(NOSY, LOCK_EX);
		seek(NOSY, 0, 2);
		print(NOSY join("\x00", ($value_pos, $container_size, $input, @whitespace))."\x01\x00");
		close(NOSY);
	}

	my @items = reverse sort { $a->[0] <=> $b->[0] }
		map { makeItem($_, $regexp, $value_pos) }  split(/\n/, $input);

	my @containers;
	my @too_large;

	# Work will be done here.
	foreach my $item (@items)
	{
		my $done = 0;
		foreach my $container (@containers)
		{
			$done && next;
			if (getContainerValue($container) + $item->[0] <= $container_size)
			{
				$done = 1;
				push(@{$container}, $item);
			}
		}

		if (!$done)
		{
			($item->[0] <= $container_size)
				? push(@containers, [$item])
				: push(@too_large, $item);
		}
	}

	# Work is done now. Generate output.
	scalar(@containers) && ($result = "\n\t\t<h2>Containers</h2>\n\t\t<ol>");
	foreach my $container (@containers)
	{
		$result .= "\n\t\t\t<li>\n\t\t\t\tSpace left in this container: ".
			($container_size - getContainerValue($container))."\n\t\t\t\t<ul>";

		foreach my $item (@{$container})
		{
			$result .= "\n\t\t\t\t\t<li>".$item->[1]."</li>";
		}

		$result .= "\n\t\t\t\t</ul>\n\t\t\t</li>";
	}
	scalar(@containers) && ($result .= "\n\t\t</ol>");

	scalar(@too_large)
		&& ($result .= "\n\t\t<h2>Too large for container size</h2>\n\t\t<ul>");
	foreach my $item (@too_large)
	{
		$result .= "\n\t\t\t<li>".$item->[1]."</li>";
	}
	scalar(@too_large) && ($result .= "\n\t\t</ul>");

	return($result, makeNumber($value_pos), makeNumber($container_size),
		quoteCode($input), @whitespace)
}
